<?php

/**
 * @file
 * Drush commands.
 *
 * Tools to make easier the use of the entity display system over the standard
 * by Drupal.
 */

/**
 * Implements hook_drush_command().
 */
function entity_display_drush_command() {

  // Export field formatters configuration.
  $items['entity-display-field-formatters-definition'] = array(
    'description' => 'Export field formatter definitions field_formatter files in the theme.',
    'callback' => '_entity_display_drush_field_formatters_export',
    'aliases' => array('ed-ffd'),
    'options' => array(
      'theme'      => t('Name of the theme where the configuration will be exported.'),
      'formatters' => t('Restrict the set of created field formatters by formatter module and type - encoded JSON.'),
      'override'   => t('Override the existing file (y-Yes or n-No). Default to Yes.'),
      'screen'     => t('Print result in screen (y-Yes or n-No). Default to Yes.'),
    ),
    'examples' => array(
      'drush ed-ffd --formatters=\'{"text":{"text_summary_or_trimmed":1,"text_trimmed":1},"date":1}\'',
    ),
  );

  // Export entity display configuration.
  $items['entity-display-entity-display'] = array(
    'description' => 'Export the entity display configuration to entity_display files in the theme.',
    'callback' => '_entity_display_drush_export',
    'aliases' => array('ed-ed'),
    'options' => array(
      'theme'    => t('Name of the theme where the entity display definitions are created.'),
      'displays' => t('Restrict the set of created displays by entity type/view mode/bundle - encoded JSON.'),
      'override' => t('Override the existing file (y-Yes or n-No). Default to Yes.'),
      'screen'   => t('Print result in screen (y-Yes or n-No). Default to Yes.'),
    ),
    'examples' => array(
      'drush ed-ed --displays=\'{"node":{"default":{"resource":1,"webinar":1}}}\'',
    ),
  );

  // Remove the existing standard display configuration in DB.
  $items['entity-display-remove-db-display'] = array(
    'description' => 'Remove the existing entity display configuration in DB. It set all the fields of a display to "Hidden".',
    'callback' => '_entity_display_drush_remove_db_display',
    'aliases' => array('ed-rdd'),
    'options' => array(
      // Important, notice the structure of the passed displays is different,
      // (bundle/view mode swap).
      'displays' => t('Restrict the set of removed displays by entity type/BUNDLE/view mode - encoded JSON.'),
      'screen'   => t('Print result in screen (y-Yes or n-No). Default to Yes.'),
    ),
    'examples' => array(
      'drush ed-rdd --displays=\'{"node":{"resource":{"default":1,"teaser":1}}}\' --screen=n',
    ),
  );

  // Get the display of a field in DB (Drupal standard).
  $items['entity-display-get-field-display'] = array(
    'description' => 'Remove the existing entity display configuration in DB. It set all the fields of a display to "Hidden".',
    'callback' => '_entity_display_drush_get_field_display',
    'aliases' => array('ed-gfd'),
    'arguments' => array(
      'field_name'  => t('Load the field configuration for the specific instance: type of the entity.'),
      'entity_type' => t('Load the field configuration for the specific instance: type of the entity.'),
      'bundle'      => t('Load the field configuration for the specific instance: entity bundle.'),
      'view_mode'   => t('Load the field configuration for the specific instance: view mode.'),
    ),
    'examples' => array(
      'drush ed-gfd title_field',
      'drush ed-gfd title_field node article',
      'drush ed-gfd title_field node article full',
    ),
  );

  return $items;
}

/**
 * Implements hook_drush_cache_clear().
 */
function entity_display_drush_cache_clear(&$types) {
  $types['entity-display'] = '_entity_display_cache_clear';
}

/**
 * Drush command callback function.
 *
 * Command: entity-display-field-formatters-definition.
 *
 * The goal of this script is to export the existing field formatter
 * configuration (or how fields are displayed) from the standard Drupal "Manage
 * Display UI" stored in DB/features to files, using the custom system
 * entity_display.
 *
 * The files are located in the project theme, in the defined for this
 * purpose location -> <theme_root>/display_configuration/
 * element_display/field_formatters. Formatters are grouped by
 * "formatter type" arrays / "formatter module" files.
 */
function _entity_display_drush_field_formatters_export() {
  // Get Drush command options.
  $options = _entity_display_get_drush_options();

  // Get the information about the existing field formatters.
  $field_formatters = array();
  $field_types = _field_info_collate_types();
  $formatters_info = $field_types['formatter types'];

  // Get the formatter settings used for each field instance. This will vary
  // by view mode.
  $field_instances = db_query("SELECT * FROM field_config_instance");
  foreach ($field_instances as $field_instance) {
    $field_info = field_info_instance($field_instance->entity_type, $field_instance->field_name, $field_instance->bundle);

    if (!empty($field_info['display'])) {
      // Get the formatter settings for the field in the corresponding
      // entity/bundle/view mode.
      foreach ($field_info['display'] as $view_mode => $display) {

        // Process only desired configurations.
        $module = $formatters_info[$display['type']]['module'];
        $type = $display['type'];
        $formatter_context = array('module' => $module, 'type' => $type);
        $is_processed_formatter = _entity_display_is_processed_configuration($options['formatters'], $formatter_context);

        // Discard hidden fields.
        if ($type != 'hidden' && $is_processed_formatter) {
          $is_new_formatter = TRUE;
          // Check if the specific formatter settings have already been stored
          // to avoid duplication. We only want to save once each formatter
          // configuration.
          if (!isset($field_formatters[$module])) {
            $field_formatters[$module] = array();
          }
          if (!isset($field_formatters[$module][$type])) {
            $field_formatters[$module][$type] = array();
          }
          else {
            // Compare the formatter with the list already saved.
            foreach ($field_formatters[$module][$type] as $formatter_key => $field_formatter) {
              if (_entity_display_equals_field_formatters($field_formatter, $display)) {
                $is_new_formatter = FALSE;
              }
            }
          }
          if ($is_new_formatter) {
            $new_formatter = array();
            $new_formatter['type'] = $type;
            $new_formatter['label'] = $display['label'];
            $new_formatter['settings'] = $display['settings'];
            // Add an index to distinguish formatter definitions belonging to
            // the same formatter type.
            $field_formatters[$module][$type][$type . '_' . count($field_formatters[$module][$type])] = $new_formatter;
          }
        }
      }
    }
  }

  // Export field formatter configuration to files.
  $summary = _entity_display_field_formatters_export_to_files($field_formatters, $options);
  // Renew the entity display cache.
  _entity_display_cache_clear();
  // Print operation summary.
  _entity_display_print_summary($summary, $options);
}

/**
 * Drush command callback function.
 *
 * Command: entity-display-entity-display.
 *
 * The goal of this script is to export the existing entity display
 * configuration (or the fields composing every entity disply) from the standard
 * Drupal "Manage Display UI" stored in DB/features to files, using the custom
 * system entity_display.
 *
 * The files are located in the project theme, in the defined for this
 * purpose location, -> <theme_root>/display_configuration/
 * entity_display. Displays are grouped by entity type/view mode in
 * folders.
 *
 * @return mixed
 *    It returns an error if the passed arguments are not correct.
 */
function _entity_display_drush_export() {
  // Clean the entity cache to ensure the created files use the up-to-date
  // field formatters configuration.
  _entity_display_cache_clear();

  // Get Drush command options.
  $options = _entity_display_get_drush_options();

  // Get the entity display configuration.
  $entity_display = array();
  $entity_info = entity_get_info();
  foreach ($entity_info as $entity_type => $entity_info) {

    // Process just entities with displays.
    if (!empty($entity_info['view modes'])) {
      foreach ($entity_info['bundles'] as $bundle => $bundle_info) {

        // Get view modes using the DEFAULT DISPLAY.
        $view_mode_settings = field_view_mode_settings($entity_type, $bundle);
        foreach ($view_mode_settings as $view_mode => $settings) {
          if (!$settings['custom_settings']) {
            $display_context = array($entity_type, $view_mode, $bundle);
            $is_processed_display = _entity_display_is_processed_configuration($options['displays'], $display_context);

            if ($is_processed_display) {
              $default_display = variable_get('entity_display_default_display', array());
              _entity_display_export_default_display_init($default_display, $entity_type, $bundle);
              $default_display[$entity_type][$bundle][$view_mode] = TRUE;
              variable_set('entity_display_default_display', $default_display);
            }
          }
        }

        // Get display configuration for NORMAL FIELDS.
        $field_instances_settings = field_info_instances($entity_type, $bundle);

        foreach ($field_instances_settings as $field_name => $field_instance) {
          foreach ($field_instance['display'] as $view_mode => $display) {

            if ($display['type'] !== 'hidden') {
              // Process only desired entity configurations.
              $display_context = array($entity_type, $view_mode, $bundle);
              $is_processed_display = _entity_display_is_processed_configuration($options['displays'], $display_context);

              if ($is_processed_display) {
                // Init all levels for the entity display configuration
                // structure.
                _entity_display_export_view_mode_init($entity_display, $entity_type, $view_mode, $bundle);

                // FIRST, save the display configuration for every bundle
                // using the view mode.
                //
                // Get the name given to a specific field formatter settings
                // in the entity_display system.
                $formatter_name = _entity_display_field_formatter_name($display);
                $formatter_settings = array(
                  'type' => $display['type'],
                  'label' => $display['label'],
                  'settings' => $display['settings'],
                  'formatter_name' => $formatter_name,
                );
                $entity_display[$entity_type][$view_mode]['bundle_displays'][$bundle][$field_name] = $formatter_settings;

                // SECOND, save the shared view mode display configuration
                // (fields using the same display for every existing bundle).
                _entity_display_export_view_mode_display_add($entity_display[$entity_type][$view_mode]['view_mode_display'], $field_name, 'field', $formatter_settings);
              }
            }
          }
        }

        // Get display configuration for EXTRA FIELDS. Unluckily, Drupal don't
        // get a method to get normal and extra fields display all together,
        // so the same (similar) process must be followed again.
        $extra_fields_settings = field_info_extra_fields($entity_type, $bundle, 'display');
        foreach ($extra_fields_settings as $extra_field => $extra_field_view_modes) {

          foreach ($extra_field_view_modes['display'] as $view_mode => $extra_field_display) {
            // field_info_extra_fields don't manage properly view modes using
            // the default display, discard them before being processed.
            if ($extra_field_display['visible'] && !$default_display[$entity_type][$bundle][$view_mode]) {
              // Process only desired entity configurations.
              $display_context = array($entity_type, $view_mode, $bundle);
              $is_processed_display = _entity_display_is_processed_configuration($options['displays'], $display_context);

              if ($is_processed_display) {
                // Init display structure.
                _entity_display_export_view_mode_init($entity_display, $entity_type, $view_mode, $bundle);
                // Save display configuration by bundle and by view mode.
                $entity_display[$entity_type][$view_mode]['bundle_displays'][$bundle][$extra_field]['formatter_name'] = 'entity_display_extra_field';
                _entity_display_export_view_mode_display_add($entity_display[$entity_type][$view_mode]['view_mode_display'], $extra_field, 'extra_field');
              }
            }
          }

        }
      }
    }
  }

  // Now we have the display for every bundle, extract the shared configuration
  // by view mode.
  foreach ($entity_display as $entity_type => $entity_view_modes) {
    foreach ($entity_view_modes as $view_mode => $view_mode_displays) {
      foreach ($view_mode_displays['view_mode_display'] as $field => $formatter_settings) {

        // Remove "shared" configurations already discarded.
        if ($formatter_settings == -1) {
          unset($entity_display[$entity_type][$view_mode]['view_mode_display'][$field]);
        }

        // It'll be considered as "shared" when it's visible and use the same
        // configuration for every bundle using the view mode.
        if ($formatter_settings != -1) {
          if ($formatter_settings['count'] == count($entity_display[$entity_type][$view_mode]['bundle_displays'])) {
            foreach ($entity_display[$entity_type][$view_mode]['bundle_displays'] as $bundle => $bundle_fields) {
              unset($entity_display[$entity_type][$view_mode]['bundle_displays'][$bundle][$field]);
            }
          }
          else {
            unset($entity_display[$entity_type][$view_mode]['view_mode_display'][$field]);
          }
        }
      }
    }
  }

  // Export the display configuration to files.
  $summary = _entity_display_export_to_files($entity_display, $options);
  // Refresh the entity display cache.
  _entity_display_cache_clear();
  // Print operation summary.
  _entity_display_print_summary($summary, $options);
}

/**
 * Drush command callback function.
 *
 * Command: entity-display-remove-db-display.
 *
 * The goal of this script is to remove any display configuration stored in DB
 * (Manage display UI) to, eventually, use the display configuration defined by
 * code through the entity display.
 */
function _entity_display_drush_remove_db_display() {
  // Get Drush command options.
  $options = _entity_display_get_drush_options();

  // Remove fields from display.
  $entity_info = entity_get_info();
  $displays_count = $fields_count = 0;
  foreach ($entity_info as $entity_type => $entity_info) {
    foreach ($entity_info['bundles'] as $bundle => $bundle_info) {
      // We pre-check if the display might belong to one of the processed
      // list as we can save time skipping unwanted configurations.
      $configuration = array($entity_type, $bundle);
      $might_be_processed_display = _entity_display_is_processed_configuration($options['displays'], $configuration, TRUE);

      if ($might_be_processed_display) {
        // Count (just once) the number of affected view modes.
        $view_modes_count = 0;
        $view_modes_counted = FALSE;

        // Print operation results.
        if ($options['screen']) {
          $complete = "\n\nPROCESSING - ENTITY TYPE: {$entity_type}, BUNDLE: {$bundle}";
          $complete .= "\n---";
          drush_log($complete, 'ok');
        }

        // NORMAL FIELDS.
        $field_instances = field_info_instances($entity_type, $bundle);
        foreach ($field_instances as $field_instance) {
          foreach ($field_instance['display'] as $view_mode => $field_display_settings) {
            $configuration = array($entity_type, $bundle, $view_mode);
            $is_processed_display = _entity_display_is_processed_configuration($options['displays'], $configuration);

            // Process only fields in the desired set of displays.
            if ($is_processed_display) {
              if (!$view_modes_counted) {
                $view_modes_count++;
              }
              // Remove the field display.
              if ($field_instance['display'][$view_mode]['type'] != 'hidden') {
                $field_instance['display'][$view_mode]['type'] = 'hidden';
                $fields_count++;

                // Print operation results.
                if ($options['screen']) {
                  $complete = "\nHidden (normal) field: {$field_instance['field_name']} - View mode: {$view_mode}";
                  drush_log($complete, 'ok');
                }
              }
            }
          }
          field_update_instance($field_instance);
          $view_modes_counted = TRUE;
        }
        // EXTRA FIELDS.
        //
        // Get the display settings for every view mode in the bundle to update
        // them.
        $view_mode_settings = field_bundle_settings($entity_type, $bundle);
        // This workaround is needed as, sometimes, field_bundle_settings don't
        // contain every extra field settings. So, let's get that info from
        // field_info_extra_fields and assign it back to field_bundle_settings.
        $display_settings = field_info_extra_fields($entity_type, $bundle, 'display');

        foreach ($display_settings as $extra_field => $extra_field_view_modes) {
          foreach ($extra_field_view_modes['display'] as $view_mode => $extra_field_display_settings) {
            $configuration = array($entity_type, $bundle, $view_mode);
            $is_processed_display = _entity_display_is_processed_configuration($options['displays'], $configuration);

            // Process only fields are visible in one of the set displays.
            if ($is_processed_display) {
              if (!$view_modes_counted) {
                $view_modes_count++;
              }
              // Remove the field display.
              if ($display_settings[$extra_field]['display'][$view_mode]['visible']) {
                $display_settings[$extra_field]['display'][$view_mode]['visible'] = FALSE;
                $fields_count++;

                // Print operation results.
                if ($options['screen']) {
                  $complete = "\nHidden (extra) field: {$field_instance['field_name']} - View mode: {$view_mode}";
                  drush_log($complete, 'ok');
                }
              }
            }
          }
          $view_mode_settings['extra_fields']['display'][$extra_field] = $display_settings[$extra_field]['display'];
          $view_modes_counted = TRUE;
        }
        field_bundle_settings($entity_type, $bundle, $view_mode_settings);
        $displays_count += $view_modes_count;
      }
    }
  }

  // Print operation summary.
  $summary = array('displays' => $displays_count, 'fields' => $fields_count);
  _entity_display_print_summary($summary, $options, 'db-clean');
}

/**
 * Drush command callback.
 *
 * Get the display settings (formatter configuration) for a field in the
 * standard Drupal display system (DB).
 *
 * If the formatter configuration has already been defined by the
 * entity_display system, get the given name. If not, returns the configuration
 * array that can be exported to it.
 *
 * @param string $field_name
 *   Name of the field (required).
 * @param string $entity_type
 *   Load the field configuration for the specific instance: type of the entity.
 * @param string $bundle
 *   Load the field configuration for the specific instance: entity bundle.
 * @param string $view_mode
 *   Load the field configuration for the specific instance: view mode.
 */
function _entity_display_drush_get_field_display($field_name, $entity_type = '', $bundle = '', $view_mode = '') {
  if (!isset($field_name)) {
    drush_log("The argument 'field_name' is required.\nCommand use: 'drush sed-gfd <field_name>* <entity_type> <bundle> <view_mode>'", 'error');
    return;
  }
  $instances_conditions = $query_variables = array();
  $fields = array('entity_type', 'bundle', 'field_name');

  foreach (drupal_map_assoc($fields) as $field) {
    if (${$field}) {
      array_push($instances_conditions, "$field = :$field");
      $query_variables[":$field"] = ${$field};
    }
    else {
      unset($fields[$field]);
    }
  }

  $query  = 'SELECT * FROM field_config_instance';
  $query .= ($instances_conditions) ? ' WHERE ' . implode(' AND ', $instances_conditions) : '';
  $query .= ($fields) ? ' ORDER BY ' . implode(',', $fields) : '';
  $field_instances = db_query($query, $query_variables);

  $complete  = "\n -----------------------------";
  $complete .= "\n ---- FIELD: {$field_name} ----";
  $complete .= "\n -----------------------------";
  drush_log($complete, 'ok');

  foreach ($field_instances as $field_instance) {
    $field_instance_data = unserialize($field_instance->data);
    $displays = $field_instance_data['display'];

    if ($view_mode) {
      $displays = array($view_mode => $displays[$view_mode]);
    }

    foreach ($displays as $instance_view_mode => $display_settings) {
      $complete = "\nENTITY TYPE: {$field_instance->entity_type} | BUNDLE: {$field_instance->bundle} | VIEW MODE: $instance_view_mode";
      // Figure out if we are currently using the field formatter settings
      // in the entity display system so that it can be reused.
      if ($formatter_name = _entity_display_field_formatter_name($displays[$instance_view_mode])) {
        $formatter_type = 'existing';
        $value = $formatter_name;
      }
      else {
        // Get just the needed info.
        $formatter_settings = array(
          'type'     => $displays[$instance_view_mode]['type'],
          'settings' => $displays[$instance_view_mode]['settings'],
          'label'    => $displays[$instance_view_mode]['label'],
        );
        $formatter_type = 'new';
        $value = "\n" . _entity_display_array_export($formatter_settings);

      }
      $complete .= "\nFORMATTER: $formatter_type";
      $complete .= "\nVALUE: $value";
      drush_log($complete, 'ok');
    }
  }
}

/**
 * Field formatters export - helper function.
 *
 * @param array $field_formatters
 *   Field formatters structure.
 * @param array $drush_options
 *   Drush command options.
 *
 * @return array
 *   Operation summary.
 */
function _entity_display_field_formatters_export_to_files($field_formatters, $drush_options) {
  $files_count = $functions_count = 0;

  // Export the field formatter definitions to the theme.
  $theme_path = drupal_get_path('theme', $drush_options['theme']);
  $formatter_folder = $theme_path . '/display_configuration/element_display/field_formatters/';
  $display_folders = array(
    $theme_path . '/display_configuration',
    $theme_path . '/display_configuration/element_display/',
    $theme_path . '/display_configuration/element_display/field_formatters/',
  );
  _entity_display_create_folder($display_folders);
  foreach ($field_formatters as $module => $formatter_types) {
    // Create the definition file.
    $file_name = DisplayDefinition::normalizeFileSuggestion($module);
    $file_name = $formatter_folder . "{$file_name}.formatter";
    if (!$drush_options['override'] && file_exists($file_name)) {
      continue;
    }

    $definition_file = fopen($file_name, 'w');
    $definition_output = '';
    $complete = '';
    $tab = '  ';

    // Generate the formatter definition output to be exported.
    $definition_output .= "<?php";
    $definition_output .= "\n\n/**";
    $definition_output .= "\n * @file";
    $definition_output .= "\n * Field formatter definition: $module.";
    $definition_output .= "\n */\n";

    foreach ($formatter_types as $formatter_type => $formatter_instances) {
      $function_name = "{$drush_options['theme']}_field_formatter_{$formatter_type}";
      $definition_output .= "\n/**";
      $definition_output .= "\n * Implements theme_field_formatter_SUGGESTION.";
      $definition_output .= "\n *";
      $definition_output .= "\n * Formatter definitions for:";
      $definition_output .= "\n *  - Formatter module: $module.";
      $definition_output .= "\n *  - Formatter type: $formatter_type.";
      $definition_output .= "\n */";
      $definition_output .= "\nfunction $function_name(" . '&$formatters) {';

      foreach ($formatter_instances as $formatter_instance => $formatter_instance_settings) {
        $definition_output .= "\n$tab" . '$formatters[\'' . $drush_options['theme'] . '_' . $formatter_instance . '\'] = ';
        // Fix coding standard issues.
        $definition_output .= _entity_display_array_export($formatter_instance_settings);
        $definition_output .= ";\n";
      }
      $definition_output .= "}\n";
      if ($drush_options['screen']) {
        $complete .= "\nImplemented function: {$function_name}";
        $functions_count++;
      }
    }

    // Place the formatter definition files in the corresponding theme folder.
    fwrite($definition_file, $definition_output);
    fclose($definition_file);
    if ($drush_options['screen']) {
      $complete = "\n---" . $complete;
      $complete = "\nGENERATED FILE: {$file_name}" . $complete;
      $complete = "\n\n---" . $complete;
      drush_log($complete, 'ok');
      $files_count++;
    }
  }

  return array('files' => $files_count, 'functions' => $functions_count);
}

/**
 * Entity display export - helper function.
 *
 * Init the structure containing view modes using the default display.
 *
 * @param array $default_display
 *   Default display structure.
 * @param string $entity_type
 *   The type of the entity to set.
 * @param string $bundle
 *   The entity bundle to set.
 */
function _entity_display_export_default_display_init(&$default_display, $entity_type, $bundle) {
  if (!isset($default_display[$entity_type])) {
    $default_display[$entity_type] = array();
  }
  if (!isset($default_display[$entity_type][$bundle])) {
    $default_display[$entity_type][$bundle] = array();
  }
}

/**
 * Entity display export - helper function.
 *
 * Init the different levels (entity type/view mode/bundle) of the structure
 * containing the entity display configuration.
 *
 * @param array $entity_display
 *   Display configuration structure.
 * @param string $entity_type
 *   The type of the entity.
 * @param string $bundle
 *   The entity bundle.
 */
function _entity_display_export_view_mode_init(&$entity_display, $entity_type, $view_mode, $bundle) {
  if (!isset($entity_display[$entity_type])) {
    $entity_display[$entity_type] = array();
  }

  if (!isset($entity_display[$entity_type][$view_mode])) {
    $entity_display[$entity_type][$view_mode] = array(
      'view_mode_display' => array(),
      'bundle_displays' => array(),
    );
  }
  if (!isset($entity_display[$entity_type][$view_mode]['bundle_displays'][$bundle])) {
    $entity_display[$entity_type][$view_mode]['bundle_displays'][$bundle] = array();
  }
}

/**
 * Entity display export - helper function.
 *
 * Add a new field configuration to a view mode display ("shared"
 * configuration).
 *
 * @param array $view_mode_display
 *   Structure containing the current view mode display.
 * @param string $field_name
 *   Name of the field.
 * @param string $type
 *   Field type ('field' or 'extra_field').
 * @param array $formatter_settings
 *   Field formatter settings.
 */
function _entity_display_export_view_mode_display_add(&$view_mode_display, $field_name, $type, $formatter_settings = array()) {
  if ($type == 'field') {
    if (!isset($view_mode_display[$field_name])) {
      $view_mode_display[$field_name] = $formatter_settings;
      $view_mode_display[$field_name]['count'] = 1;
    }
    // The field setting is discarded (-1) when there is a single bundle
    // using different formatter settings for the field.
    elseif ($view_mode_display[$field_name] != -1) {
      if (!_entity_display_equals_field_formatters($view_mode_display[$field_name], $formatter_settings)) {
        $view_mode_display[$field_name] = -1;
      }
      // We'll need to know later how many bundles are using the same formatter.
      else {
        $view_mode_display[$field_name]['count']++;
      }
    }
  }
  elseif ($type == 'extra_field') {
    if (!isset($view_mode_display[$field_name])) {
      $view_mode_display[$field_name]['formatter_name'] = 'entity_display_extra_field';
      $view_mode_display[$field_name]['count'] = 1;
    }
    else {
      $view_mode_display[$field_name]['count']++;
    }
  }
}

/**
 * Entity display export - helper function.
 *
 * Export the display configuration to files.
 *
 * @param array $entity_display
 *   Display configuration structure.
 * @param array $options
 *   Drush command options.
 *
 * @return array
 *   Operation summay.
 */
function _entity_display_export_to_files($entity_display, $options) {
  $files_count = $functions_count = 0;
  $theme = $options['theme'];

  // Export the field formatter definitions the theme.
  $theme_path = drupal_get_path('theme', $theme);
  $display_folders = array(
    $theme_path . '/display_configuration',
    $theme_path . '/display_configuration/entity_display/',
  );
  _entity_display_create_folder($display_folders);
  $formatter_folder = $theme_path . '/display_configuration/entity_display/';
  foreach ($entity_display as $entity_type => $view_modes) {
    // Create a folder to group all the displays for the entity.
    $entity_folder = $formatter_folder . "$entity_type";
    _entity_display_create_folder(array($entity_folder));
    foreach ($view_modes as $view_mode => $view_mode_formatters) {
      // Create a folder to group all the displays for the view mode.
      $view_mode_folder = $entity_folder . "/$view_mode";
      _entity_display_create_folder(array($view_mode_folder));

      // Set file data.
      $entity_data = array(
        'entity_type' => $entity_type,
        'view_mode'   => $view_mode,
      );
      $file_context = array(
        'theme' => $theme,
        'file_folder' => $view_mode_folder,
      );

      // Create the definition files with the specific by bundle display.
      foreach ($view_mode_formatters['bundle_displays'] as $bundle => $bundle_fields) {
        // Customize data before creating the file.
        $entity_data['bundles'] = array($bundle);
        $file_context['limited_displays'] = TRUE;
        $updates_count    = _entity_display_export_generate_file($bundle_fields, $entity_data, $file_context, $options);
        $files_count     += $updates_count['files'];
        $functions_count += $updates_count['functions'];
      }

      // Create the file with the entity display definition shared by every
      // bundle.
      $entity_data['bundles'] = array_keys($view_mode_formatters['bundle_displays']);
      $file_context['limited_displays'] = !empty($options['displays']) ? TRUE : FALSE;
      $updates_count    = _entity_display_export_generate_file($view_mode_formatters['view_mode_display'], $entity_data, $file_context, $options);
      $files_count     += $updates_count['files'];
      $functions_count += $updates_count['functions'];
    }
  }

  return array('files' => $files_count, 'functions' => $functions_count);
}

/**
 * Entity display export - helper function.
 *
 * Generate an entity display definition file.
 *
 * @param array $field_formatters
 *   List of fields (keys) and field formatter names (values) composing the
 *   display.
 * @param array $entity_data
 *    Context information of the entity display created:
 *      - $context['entity_type'].
 *      - $context['view_mode'].
 *      - $context['bundles'].
 * @param array $file_context
 *    Data needed to create the definition file:
 *      - $file_context['theme']: Theme where the displays will be created.
 *      - $file_context['file_folder']: Folder path where the display definition
 *        file will be created.
 *      - $file_context['limited_displays']: TRUE if the display is create by a
 *        reduced number of bundles of a view mode. FALSE, if the display
 *        belongs to whole view mode.
 * @param array $drush_options
 *    Contains the entered drush command options.
 *
 * @return array
 *   Elements (files and functions) generated count.
 */
function _entity_display_export_generate_file($field_formatters, $entity_data, $file_context, $drush_options) {
  // Init updates count.
  $updates_count = array(
    'files' => 0,
    'functions' => 0,
  );
  // Get entity data.
  $view_mode   = $entity_data['view_mode'];
  $entity_type = $entity_data['entity_type'];
  $bundles     = $entity_data['bundles'];
  // Get context data.
  $limited_displays = $file_context['limited_displays'];
  $folder           = $file_context['file_folder'];
  $theme            = $file_context['theme'];

  // Create the definition file.
  $group_name = $limited_displays ? '--' . implode($bundles, '--') : '';
  $file_name = $folder . "/" . DisplayDefinition::normalizeFileSuggestion("{$entity_type}--{$view_mode}{$group_name}.display");

  // Skip the creation of the file:
  // - If the display definition file exists and the user selected the option
  //   to not override existing configurations, skip the creation of the file.
  // - There is no fields to include in the entity display.
  if ((file_exists($file_name) && !$drush_options['override']) || empty($field_formatters)) {
    return $updates_count;
  }

  $definition_file = fopen($file_name, 'w');
  $tab = '  ';
  $complete = '';
  $definition_output = '';
  $group_name = $limited_displays ? '_' . implode($bundles, '_') : '';
  $function_name = "{$theme}_entity_display_{$entity_type}_{$view_mode}{$group_name}";

  // Generate the formatter definition output to be exported.
  $definition_output .= "<?php";
  $definition_output .= "\n\n/**";
  $definition_output .= "\n * @file";
  $definition_output .= "\n * Entity display definition.";
  $definition_output .= "\n *";
  $definition_output .= "\n * Entity type: $entity_type - View mode: $view_mode";
  $definition_output .= $limited_displays ? "- Bundle/s: " . implode(', ', $bundles) . "." : ".";
  $definition_output .= "\n */\n";

  $definition_output .= "\n/**";
  $definition_output .= "\n * Implements theme_entity_display_SUGGESTION.";
  $definition_output .= "\n */";
  $definition_output .= "\nfunction {$function_name}(" . '&$display_elements, $entity_type, $bundle, $view_mode) {';

  $condition = '($entity_type == ' . "'{$entity_type}'" . ' && $view_mode == ' . "'{$view_mode}')";
  if ($limited_displays) {
    $bundle_conditions = array();
    foreach ($bundles as $bundle) {
      $bundle_condition = '$bundle == ' . "'{$bundle}'";
      array_push($bundle_conditions, $bundle_condition);
    }
    $bundle_conditions = implode(' || ', $bundle_conditions);
    $condition .= " && ({$bundle_conditions})";
  }
  $definition_output .= "\n\n{$tab}if ({$condition}) {";

  foreach ($field_formatters as $field_name => $field_formatter) {
    $definition_output .= "\n{$tab}{$tab}" . '$display_elements[\'' . $field_name . '\'] = \'' . $field_formatter['formatter_name'] . '\';';
  }
  $definition_output .= "\n{$tab}}\n";
  $definition_output .= "}\n";
  // Print in screen.
  $updates_count['functions']++;
  if ($drush_options['screen']) {
    $complete .= "\nImplemented function: {$function_name}";
  }

  // Place the formatter definition files in the corresponding theme folder.
  fwrite($definition_file, $definition_output);
  fclose($definition_file);
  // Print in screen.
  $updates_count['files']++;
  if ($drush_options['screen']) {
    $complete = "\n---" . $complete;
    $complete = "\nGENERATED FILE: {$file_name}" . $complete;
    $complete = "\n\n---" . $complete;
    drush_log($complete, 'ok');
  }

  return $updates_count;
}

/**
 * Util function - Get Drush command options.
 *
 * @return array
 *   Used command options.
 */
function _entity_display_get_drush_options() {
  global $theme, $base_theme_info;

  // Restricted formatters. Only these formatter module/type will be processed.
  $options['formatters'] = drush_get_option('formatters');
  $options['formatters'] = json_decode($options['formatters'], TRUE);

  // Restricted displays. Only the entity_type, view mode and bundle
  // configurations included here will be processed.
  $options['displays'] = drush_get_option('displays');
  $options['displays'] = json_decode($options['displays'], TRUE);

  // Override existing files.
  $options['override'] = drush_get_option('override', TRUE);
  $options['override'] = ($options['override'] === 'n') ? FALSE : TRUE;

  // Print results in screen.
  $options['screen'] = drush_get_option('screen', TRUE);
  $options['screen'] = ($options['screen'] === 'n') ? FALSE : TRUE;

  // Theme.
  $theme_base = drush_get_option('theme', FALSE);
  // If a theme was passed as an option, check its value is correct.
  if ($theme_base) {
    $theme_keys = array();
    foreach ($base_theme_info as $base) {
      $theme_keys[] = $base->name;
    }
    $theme_keys[] = $theme;

    if (!in_array($theme, $theme_keys)) {
      return drush_set_error(dt('"@theme" is not a valid theme name. Please, review the passed argument.', array('@theme' => $theme_base)));
    }
  }
  // If no theme is chosen, use the theme base.
  else {
    $theme_info = reset($base_theme_info);
    $theme_base = $theme_info->name;
  }
  $options['theme'] = $theme_base;

  return $options;
}

/**
 * Util function - Get the name of a field formatter in the Display system.
 *
 * Given a field formatter definition it retrieves the name given when defined
 * through a hook_field_formatter().
 *
 * @param string $field_formatter
 *   Field formatter definition.
 *
 * @return mixed
 *   The name of the formatter or NULL if it is not found.
 */
function _entity_display_field_formatter_name($field_formatter) {
  $entity_display = new EntityDisplay();
  $entity_display->initRenderElements();
  $render_elements = $entity_display->getConfElements();

  foreach ($render_elements['field'] as $formatter_name => $formatter_settings) {
    if (_entity_display_equals_field_formatters($formatter_settings, $field_formatter)) {
      return $formatter_name;
    }
  }
  return NULL;
}

/**
 * Util function - Check if two field formatter settings are equal.
 *
 * Elements checked: 'type', 'settings' and 'label'.
 *
 * @param array $field_formatter_1
 *   First field formatter to compare.
 * @param array $field_formatter_2
 *   Second field formatter to compare.
 *
 * @return bool
 *    TRUE if they are the same field formatter. FALSE, otherwise.
 */
function _entity_display_equals_field_formatters($field_formatter_1, $field_formatter_2) {
  if ($field_formatter_1['type'] == $field_formatter_2['type']) {
    if ($field_formatter_1['label'] == $field_formatter_2['label']) {
      if (empty($field_formatter_1['settings']) && empty($field_formatter_2['settings'])) {
        return TRUE;
      }
      $array_diff = array_diff($field_formatter_1['settings'], $field_formatter_2['settings']);
      if (empty($array_diff)) {
        return TRUE;
      }
    }
  }
  return FALSE;
}

/**
 * Util function - Check if a configuration is set up to be processed.
 *
 * Helper for drush commands to process only a specific set of display elements.
 *
 * @param array $processed_configurations
 *   List of configurations to be processed.
 * @param array $configuration
 *   Configuration to check.
 * @param bool $partial
 *   FALSE, if it is checking completely the processed configurations (until the
 *   last elements, which must be equal to "1"), or TRUE if it is checking the
 *   passed configuration is partially contained.
 *
 * @return bool
 *   TRUE if $configuration is part of the configurations
 *   ($processed_configurations) to be processed.
 */
function _entity_display_is_processed_configuration($processed_configurations, $configuration, $partial = FALSE) {
  if (empty($processed_configurations)) {
    return TRUE;
  }

  foreach ($configuration as $index => $key) {
    if (isset($processed_configurations[$key])) {
      if ($partial && (count($configuration) == $index + 1)) {
        return TRUE;
      }
      if ($processed_configurations[$key] === 1) {
        return TRUE;
      }
      elseif (is_array($processed_configurations[$key])) {
        $processed_configurations = &$processed_configurations[$key];
      }
      else {
        return FALSE;
      }
    }
    else {
      return FALSE;

    }
  }
  return FALSE;
}

/**
 * Util function - Create a list of folders if they don't exist.
 *
 * @param array $folders
 *    List of folder paths to create.
 *
 * @return array
 *   Summary of the operation.
 */
function _entity_display_create_folder($folders) {
  $errors = array();
  foreach ($folders as $folder) {
    if (!file_exists($folder)) {
      if (!mkdir($folder)) {
        $errors[] = "Creating the folder $folder";
      }
    }
  }

  return array('errors' => $errors);
}

/**
 * Util function - Print the results of an operation.
 *
 * @param array $summary
 *   Operation summay.
 * @param array $drush_options
 *   Drush command options.
 */
function _entity_display_print_summary($summary, $drush_options, $type = 'export') {
  if ($drush_options['screen']) {
    // Operation summary.
    $complete = "\n\n--------------- PROCESSING COMPLETE ---------------\n";

    if ($type == 'export') {
      // Operation summary.
      $complete = "\n\n--------------- PROCESSING COMPLETE ---------------\n";
      $complete .= "THEME: {$drush_options['theme']}\n";
      $complete .= "FILES GENERATED: {$summary['files']}\n";
      $complete .= "IMPLEMENTED FUNCTIONS: {$summary['functions']}\n\n";
      drush_log($complete, 'ok');
    }
    elseif ($type == 'db-clean') {
      $complete .= "DISPLAYS AFFECTED: {$summary['displays']}\n";
      $complete .= "HIDDEN FIELDS: {$summary['fields']}\n\n";
      drush_log($complete, 'ok');
    }
  }
}

/**
 * Util function - string representation of an array...
 *
 * ...following Drupal coding standards.
 *
 * - The string will be ready to be printed in a configuration file.
 * - It supports up to three level arrays.
 *
 * @param array $variable
 *   Array to be exported.
 *
 * @return string
 *   Formatted representation of the array.
 */
function _entity_display_array_export($variable) {
  $tab = ' ';
  $export_elements_formatted = array(
    // Remove space in starting array parenthesis.
    'array ('     => 'array(',
    // Remove excess tabs.
    "\n{$tab}array(" => 'array(',
    "\n{$tab}{$tab}array(" => 'array(',
    "\n{$tab}{$tab}{$tab}array(" => 'array(',
    // Fix empty arrays.
    "array(\n{$tab})" => "array()",
    "array(\n{$tab}{$tab})" => "array()",
    "array(\n{$tab}{$tab}{$tab})" => "array()",
    // Include additional tab in the array declaration.
    "\n"          => "\n{$tab}",
    // Capitol letters for boolean variables.
    "true"        => 'TRUE',
    "false"       => 'FALSE',
  );

  return str_replace(array_keys($export_elements_formatted), $export_elements_formatted, var_export($variable, TRUE));
}
